from datetime import datetime, timedelta, timezone
from unittest import IsolatedAsyncioTestCase, TestCase
from unittest.mock import call

from fastapi import HTTPException, Request
from fastapi.security import SecurityScopes
from fastapi.testclient import TestClient
from fred_types import ContactId
from freezegun import freeze_time
from jose import jwe, jwt
from jose.constants import ALGORITHMS
from regal.exceptions import ContactDoesNotExist
from verifier import PaginatedList, VerificationId
from verifier.exceptions import VerificationDoesNotExist

from inquisitor.auth import (
    OAuth2PasswordBearerWithCookie,
    TokenData,
    auth_token_data,
    create_access_token,
    get_contact_id,
    get_verification,
    get_verification_id,
)
from inquisitor.constants import OAUTH_COOKIE, Scope
from inquisitor.main import app
from inquisitor.settings import Settings, get_settings

from .auth import TestGetAccessToken
from .utils import (
    CONTACT_ID,
    VERIFICATION_ID,
    ContactClientMixin,
    LoggerMixin,
    VerificationClientMixin,
    make_verification,
)

client = TestClient(app)


class OAuth2PasswordBearerWithCookieTest(IsolatedAsyncioTestCase):
    def setUp(self):
        self.oauth2_scheme = OAuth2PasswordBearerWithCookie(tokenUrl="auth/token/", scopes={})

    async def test_oauth2_cookie_present(self):
        request = Request(scope={"type": "http", "headers": {}})
        request.cookies[OAUTH_COOKIE] = "Bearer Gazpacho!"
        self.assertEqual(await self.oauth2_scheme(request), "Gazpacho!")

    async def test_oauth2_cookie_missing(self):
        request = Request(scope={"type": "http", "headers": {}})
        with self.assertRaises(HTTPException) as exc:
            await self.oauth2_scheme(request)
        self.assertEqual(exc.exception.status_code, 401)
        self.assertEqual(exc.exception.detail, "Not authenticated")
        self.assertEqual(exc.exception.headers, {"WWW-Authenticate": "Bearer"})


class CreateAccessTokenTest(TestCase):
    @freeze_time("2022-01-01T00:00:00Z")
    def test_create_access_token(self):
        settings = Settings()
        settings.secret_key = "Gazpacho!"
        settings.access_token_expires = int(timedelta(days=1).total_seconds())

        access_token = create_access_token(
            TokenData(
                verification_id=VerificationId("v:1"),
                contact_id=ContactId("c:1"),
                scopes=[Scope.READ],
            ),
            settings=settings,
        )

        decrypted = jwe.decrypt(access_token, settings.secret_key_sha256)
        self.assertEqual(
            jwt.decode(decrypted, settings.secret_key, algorithms=[ALGORITHMS.HS256]),
            {
                "verification_id": "v:1",
                "contact_id": "c:1",
                "scopes": [Scope.READ],
                "nbf": int((datetime.utcnow() - timedelta(minutes=5)).timestamp()),
                "iat": int(datetime.utcnow().timestamp()),
                "exp": int((datetime.utcnow() + timedelta(days=1)).timestamp()),
            },
        )


class GetAccessTokenTest(TestGetAccessToken.Base):
    url = "/auth/token/"

    def test_login_without_password(self):
        response = client.post(self.url, data={"username": "v:1"})

        self.assertEqual(response.json(), {"detail": "Password is required"})
        self.assertEqual(response.status_code, 400)
        self.assertNoLogs()


class AuthTokenDataTest(IsolatedAsyncioTestCase):
    async def test_auth_token_data(self):
        token_data = TokenData(
            verification_id=VerificationId("v:1"), contact_id=ContactId("c:1"), scopes=[Scope.READ]
        )
        access_token = create_access_token(token_data, get_settings())

        self.assertEqual(
            await auth_token_data(
                security_scopes=SecurityScopes([Scope.READ]),
                token=access_token,
                settings=get_settings(),
            ),
            token_data,
        )

    async def test_insufficient_permissions(self):
        token_data = TokenData(
            verification_id=VerificationId("v:1"), contact_id=ContactId("c:1"), scopes=[Scope.READ]
        )
        access_token = create_access_token(token_data, get_settings())

        with self.assertRaises(HTTPException) as exc:
            await auth_token_data(
                security_scopes=SecurityScopes([Scope.WRITE]),
                token=access_token,
                settings=get_settings(),
            )
        self.assertEqual(exc.exception.status_code, 403)
        self.assertEqual(exc.exception.detail, "Insufficient permissions")
        self.assertEqual(exc.exception.headers, {"WWW-Authenticate": 'Bearer scope="write"'})

    async def test_invalid_token(self):
        with self.assertRaises(HTTPException) as exc:
            await auth_token_data(
                security_scopes=SecurityScopes([Scope.WRITE]),
                token="invalid.token",
                settings=get_settings(),
            )
        self.assertEqual(exc.exception.status_code, 401)
        self.assertEqual(exc.exception.detail, "Could not validate credentials")
        self.assertEqual(exc.exception.headers, {"WWW-Authenticate": "Bearer"})


class GetContactIdTest(VerificationClientMixin, IsolatedAsyncioTestCase):
    async def test_get_contact_id(self):
        self.assertEqual(
            await get_contact_id(
                TokenData(verification_id=VerificationId("v:1"), contact_id=ContactId("c:1"))
            ),
            ContactId("c:1"),
        )


class GetVerificationIdTest(VerificationClientMixin, IsolatedAsyncioTestCase):
    async def test_get_verification_id(self):
        self.assertEqual(
            await get_verification_id(
                TokenData(verification_id=VerificationId("v:1"), contact_id=ContactId("c:1"))
            ),
            VerificationId("v:1"),
        )


class GetVerificationTest(VerificationClientMixin, IsolatedAsyncioTestCase):
    async def test_get_verification(self):
        verification = make_verification()
        self.verification_client.get.return_value = verification
        self.assertEqual(
            await get_verification(VerificationId("v:1"), self.verification_client),
            verification,
        )
        self.assertEqual(self.verification_client.mock_calls, [call.get(VerificationId("v:1"))])

    async def test_get_verification_not_found(self):
        self.verification_client.get.side_effect = VerificationDoesNotExist

        with self.assertRaises(HTTPException) as exc:
            await get_verification(VerificationId("v:1"), self.verification_client)

        self.assertEqual(exc.exception.status_code, 404)
        self.assertEqual(exc.exception.detail, "Verification not found")
        self.assertEqual(self.verification_client.mock_calls, [call.get(VerificationId("v:1"))])


@freeze_time("2022-01-01T00:00:00Z")
class GetRepresentativeAccessTokenTest(
    ContactClientMixin, VerificationClientMixin, LoggerMixin, TestCase
):
    url = "/auth/representative/token/"

    def test_login_without_password(self):
        response = client.post(self.url, data={"username": "RIMMER"})

        self.assertEqual(response.json(), {"detail": "Password is required"})
        self.assertEqual(response.status_code, 400)

    def test_wrong_contact_handle(self):
        self.contact_client.get_contact_id.side_effect = ContactDoesNotExist
        response = client.post(self.url, data={"username": "HOLLY", "password": "starbug"})

        self.assertEqual(response.json(), {"detail": "Incorrect username or password"})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.contact_client.mock_calls, [call.get_contact_id("HOLLY")])
        self.assertEqual(self.verification_client.mock_calls, [])

    def test_no_verifications(self):
        self.contact_client.get_contact_id.return_value = "c:1"
        self.verification_client.list.return_value = PaginatedList(results=[])

        response = client.post(self.url, data={"username": "RIMMER", "password": "starbug"})

        self.assertEqual(response.json(), {"detail": "Verification not found"})
        self.assertEqual(response.status_code, 404)
        self.assertEqual(self.contact_client.mock_calls, [call.get_contact_id("RIMMER")])
        self.assertEqual(self.verification_client.mock_calls, [call.list(contact_id="c:1")])

    def test_multiple_verifications(self):
        self.contact_client.get_contact_id.return_value = "c:1"
        self.verification_client.list.return_value = PaginatedList(
            results=[
                make_verification(
                    verification_id="v:1",
                    create_datetime=datetime(2022, 1, 1, tzinfo=timezone.utc),
                ),
                make_verification(
                    verification_id="v:2",
                    create_datetime=datetime(2021, 1, 1, tzinfo=timezone.utc),
                    close_datetime=datetime(2021, 2, 1, tzinfo=timezone.utc),
                ),
            ]
        )
        self.verification_client.verify_token.return_value = True
        self.verification_client.get.return_value = make_verification()

        response = client.post(self.url, data={"username": "RIMMER", "password": "starbug"})

        self.assertEqual(response.json(), {"scopes": [Scope.READ]})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(self.contact_client.mock_calls, [call.get_contact_id("RIMMER")])
        self.assertEqual(
            self.verification_client.mock_calls,
            [
                call.list(contact_id="c:1"),
                call.verify_token(VerificationId("v:1"), "starbug", check=True),
                call.get(VerificationId("v:1"), access_token=""),
            ],
        )

    def test_invalid_verify_token(self):
        self.contact_client.get_contact_id.return_value = "c:1"
        self.verification_client.list.return_value = PaginatedList(results=[make_verification()])
        self.verification_client.verify_token.return_value = False

        response = client.post(self.url, data={"username": "RIMMER", "password": "starbug"})

        self.assertEqual(response.json(), {"detail": "Incorrect username or password"})
        self.assertEqual(response.status_code, 400)
        self.assertEqual(self.contact_client.mock_calls, [call.get_contact_id("RIMMER")])
        self.assertEqual(
            self.verification_client.mock_calls,
            [
                call.list(contact_id="c:1"),
                call.verify_token(VerificationId("v:1"), "starbug", check=True),
            ],
        )


class RefreshAccessTokenTest(TestCase):
    def setUp(self):
        self.settings = Settings(secret_key="Gazpacho!", access_token_expires=24 * 60 * 60)

        app.dependency_overrides[get_settings] = lambda: self.settings
        app.dependency_overrides[auth_token_data] = lambda: TokenData(
            verification_id=VERIFICATION_ID,
            contact_id=CONTACT_ID,
            scopes=[Scope.READ, Scope.WRITE],
        )

    def tearDown(self):
        app.dependency_overrides = {}

    @freeze_time("2022-01-01T01:00:00Z")
    def test_refresh_access_token(self):
        response = client.post("/auth/token/refresh/")

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), {"scopes": [Scope.READ, Scope.WRITE]})
        access_token = response.cookies[OAUTH_COOKIE].removeprefix('"Bearer ').removesuffix('"')
        decrypted = jwe.decrypt(access_token, self.settings.secret_key_sha256)
        payload = jwt.decode(decrypted, self.settings.secret_key, algorithms=[ALGORITHMS.HS256])
        self.assertEqual(
            payload,
            {
                "verification_id": str(VERIFICATION_ID),
                "contact_id": str(CONTACT_ID),
                "exp": datetime(2022, 1, 2, 1, tzinfo=timezone.utc).timestamp(),
                "iat": datetime(2022, 1, 1, 1, tzinfo=timezone.utc).timestamp(),
                "nbf": datetime(2022, 1, 1, 0, 55, tzinfo=timezone.utc).timestamp(),
                "scopes": [Scope.READ.value, Scope.WRITE.value],
            },
        )


class LogoutTest(TestCase):
    def test_logout(self):
        client = TestClient(app, cookies={OAUTH_COOKIE: "Gazpacho!", "some": "value"})
        response = client.post("/auth/logout/")

        self.assertEqual(response.status_code, 200)
        # There is some bug in starlette that doesn't allow testing response.cookies directly.
        self.assertIn('access_token=""', response.headers["set-cookie"])
